/*
 * Author:	Brian J. Murrell <brian@interlinx.bc.ca>
 *
 * Stateful Packet Filter -- constantly adjust an inbound packet filter
 * to deal with allowing traffic in based on what was sent out.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
/*
#include <netinet/ip_tcp.h>
*/
#include <arpa/inet.h>
#include "kernel_headers.h"

struct fwhdr {
	__u32 mark;
	char interface[IFNAMSIZ];
};

#ifndef MAX
#define MAX(_a,_b)	((_a)<(_b)?(_b):(_a))
#endif
#ifndef MIN
#define MIN(_a,_b)	((_a)<(_b)?(_a):(_b))
#endif

/* this should be largest value a time_t can hold without overflowing */
#define LONG_TIMEOUT	0x7FFFFFFFL;
#define	DEFAULT_ACTION	"DENY"	
#define TRACEROUTE_MIN	33435
#define TRACEROUTE_MAX	TRACEROUTE_MIN+30*3
/* what rule to start inserting dyn. rules */
#ifdef ALLOW_EST_TCP
#define START_DYN_RULES	2	
#else
#define START_DYN_RULES	1
#endif

static time_t next_timeout;	/* absolute time of next select loop process */

struct connection {
	struct in_addr src_addr;
	struct in_addr src_mask;
	struct in_addr dst_addr;
	struct in_addr dst_mask;
	uint proto;
	uint src_port_lo;
	uint src_port_hi;
	uint dst_port_lo;
	uint dst_port_hi;
	char interface[IFNAMSIZ];
	time_t timeout;
	struct connection *prev;
	struct connection *next;
};

struct connection *connections=NULL, *last_connection, *first_connection;

uint debuglev=0;

#define add_tcp_pr(sa, da, spl, sph, dpl, dph, s, i) add_rule(sa, 0xFFFFFFFF, da, 0xFFFFFFFF, spl, sph, dpl, dph, IPPROTO_TCP, s, i)
#define add_tcp_mask(sa, sm, da, dm, sp, dp, s, i) add_rule(sa, sm, da, dm, sp, sp, dp, dp, IPPROTO_TCP, s, i)
#define add_udp_mask(sa, sm, da, dm, sp, dp, s, i) add_rule(sa, sm, da, dm, sp, sp, dp, dp, IPPROTO_UDP, s, i)
#define add_icmp_mask(sa, sm, da, dm, t, c, s, i) add_rule(sa, sm, da, dm, t, t, c, c, IPPROTO_ICMP, s, i)
#define add_tcp(sa, da, sp, dp, s, i) add_rule(sa, 0xFFFFFFFF, da, 0xFFFFFFFF, sp, sp, dp, dp, IPPROTO_TCP, s, i)
#define add_udp(sa, da, sp, dp, s, i) add_rule(sa, 0xFFFFFFFF, da, 0xFFFFFFFF, sp, sp, dp, dp, IPPROTO_UDP, s, i)
#define add_icmp(sa, da, t, c, s, i) add_rule(sa, 0xFFFFFFFF, da, 0xFFFFFFFF, t, t, c, c, IPPROTO_ICMP, s, i)

extern uint nsaddrs, ndaddrs;
extern struct in_addr *saddrs, *daddrs;
extern int flush_entries(const ip_chainlabel chain);
void hdump(register const u_char *, register u_int);
int debug(char *, ...);
struct connection *find_connection(struct in_addr, struct in_addr,
		struct in_addr, struct in_addr, uint, uint, uint, uint, uint,
		char *);
void cleanup_exit(int);

int
main(int argc, char *argv[]) {

	char packet[65555];
	int stream, c;
	ssize_t num_read;

	while ((c=getopt(argc, argv, "x:"))!=EOF)
		switch (c) {
			case 'x':
				debuglev=atoi(optarg);
				break;
			default:
				cleanup_exit(SIGTERM);
		}
	if ((stream=open("/dev/fwdata", O_RDONLY))<0) {
		perror("open /dev/fwdata");
		exit(1);
	}

	/* cleanup and exit when we get various signals */
	signal(SIGINT, cleanup_exit);
	signal(SIGTERM, cleanup_exit);
	signal(SIGSEGV, cleanup_exit);

	/* install filters needed to do stateful filtering */
	install_spf_filters();

	next_timeout=LONG_TIMEOUT;
	while(1) {
		int ret;
		fd_set readfds;
		struct timeval timeout;

		FD_ZERO(&readfds);
		FD_SET(stream, &readfds);
		timeout.tv_usec=0;
		timeout.tv_sec=next_timeout-time((time_t)NULL);
		if ((ret=select(stream+1, &readfds, NULL, NULL, &timeout))<0) {
			if (errno != EINTR) {
				perror("select");
				exit(1);
			}
		} else if (ret==0) {
			/* check the connections table for expired entries */
			expire_connections();
		} else if (ret>0) {
			if (FD_ISSET(stream, &readfds)) {
				if ((num_read=read(stream, packet,
					      sizeof(char)*2000))<-1) {
					perror("read");
					exit(1);
				}
				do_packet(packet, num_read);
			}
		}
		debug("\n");
		if (debuglev>1)
			display_connections();
	}

}

/* Flush all rules in all chains and exit */
void
cleanup_exit(int sig) {

	debug("Got signal %d, flushing chains and exiting\n", sig);
	for_each_chain(flush_entries);
	/* remove our statinpt chain */
	if (delete_chain("statinpt", 0)<0)
		perror("removing statinpt chain");
	exit(0);
}

#define BYTESTR 16
#define BYTELEN  4

void
hdump(register const u_char *bp, register u_int length)
{

/* this could probably be more elegant, but it was ripped off from "hdump"
 * without much "how does it work" questions asked
 */
	long pos;
	size_t nread;
	register const u_char *sp;
	register u_int i;
	int flen = BYTELEN;

	if (length<0) {
		/* can't print if the data size is less that the offset */
		return;
	}
#if 0
	if ((long)bp & 1) {
		default_print_unaligned(bp, length);
		return;
	}
#endif
	pos = 0;
	while ( length > 0 ) {
		if (length < BYTESTR)
			nread = length;
		else
			nread = BYTESTR;
		sp = bp + pos;

		printf ( "\n%8ld:  ", pos );

		for ( i = 0; i < BYTESTR; i++ ) {
			if ( i < nread )
				printf ( "%02x", sp[i] );
			else
				printf ( "  " );
			if ( ( i % flen ) == ( flen - 1 ) )
				printf ( " " ); 
		}

		printf ( "  " ); 
		for ( i = 0; i < BYTESTR; i++ ) {
			if ( i < nread ) {
				if ( isprint ( sp[i] ) )
					printf ( "%c", sp[i] );
				else
					printf ( "." );
			}
		}

		pos += nread;
		length -= nread;

	}
	printf("\n");
}

int
debug(char *format, ...) {

	if (debuglev>0) {
		va_list ap;

		va_start(ap, format);
		(void)vfprintf(stderr, format, ap);
		va_end(ap);
	}

}

do_packet(const char *p, int num_read) {

	char *ip_src, *ip_dst;
	struct ip *ip;

	(char *)ip=p+20;

	if (debuglev>0) {
		ip_src=strdup(inet_ntoa(ip->ip_src));
		ip_dst=strdup(inet_ntoa(ip->ip_dst));
		debug("(%d) %s > %s", num_read, ip_src, ip_dst);
		if (debuglev>1)
			hdump(p, num_read);
		free(ip_src);
		free(ip_dst);
	}

	switch(ip->ip_p) {
		case IPPROTO_TCP:
			do_tcp(p);
			break;
		case IPPROTO_UDP:
			do_udp(p);
			break;
		case IPPROTO_ICMP:
			do_icmp(p);
			break;
		default:
			fprintf(stderr, "unknown IP protocol \"%d\"\n",
				ip->ip_p);
	}

}

int
do_tcp(const char *p) {

	struct ip *ip;
	struct tcphdr *tcp;
	struct fwhdr *fw=(struct fwhdr *)p;
	char *data;

	debug(" tcp");

	(char *)ip=p+20;
	(char *)tcp=(char *)ip+ip->ip_hl*4;
	debug(" %d %d", ntohs(tcp->source), ntohs(tcp->dest));

#ifndef ALLOW_EST_TCP
	/*
	 * Allow return TCP packets
	 */
	/* install a reverse rule for 30 seconds */
	add_tcp(ip->ip_dst, ip->ip_src, ntohs(tcp->dest),
		ntohs(tcp->source), 30, fw->interface);
#endif
	
	data=(char *)tcp+tcp->doff*4;
	/* Watch for and open holes for FTP DATA sessions */
	if (ntohs(tcp->dest)==21) {
		if (debuglev>1)
			hdump(data,
			      ntohs(ip->ip_len)-(ip->ip_hl*4)-(tcp->doff*4));
		/* this is purely *cheating* cheapo code.  assuming this stuff
		 * will always be at byte 0 of a data packet is naive.  the
		 * proper solution is to keep a running buffer which we always
		 * search for what we want
		 */
		if (strncasecmp(data, "port ", 4)==0) {
			/* parse the port specification */

			unsigned short int port;
			char *c, *p;

			p=(char *)&port;
			c=strrchr(data, ',');
			*c++='\0';
			p[0]=strtol(c, (char **)NULL, 10);
			c=strrchr(data, ',');
			*c++='\0';
			p[1]=strtol(c, (char **)NULL, 10);
			add_tcp(ip->ip_dst, ip->ip_src, 20, port, 30,
				fw->interface);
		}
	} else if (ntohs(tcp->source)==21) {
		if (strncasecmp(data, "227 Entering Passive Mode", 25)==0) {
			/* parse the port specification */

			unsigned short int port;
			char *c, *p=(char *)&port;

			c=strrchr(data, ')');
			*c='\0';
			c=strrchr(data, ',');
			*c++='\0';
			p[0]=strtol(c, (char **)NULL, 10);
			c=strrchr(data, ',');
			*c++='\0';
			p[1]=strtol(c, (char **)NULL, 10);
			add_tcp_pr(ip->ip_dst, ip->ip_src, 0, 65535, port,
				   port, 30, fw->interface);
		}
	}
}

int
do_udp(const char *p) {

	struct ip *ip;
	struct udphdr *udp;
	struct fwhdr *fw=(struct fwhdr *)p;

	debug(" udp");

	(char *)ip=p+20;
	(char *)udp=(char *)ip+ip->ip_hl*4;

	/* outbound traceroutes should allow corresponding inbound "ttl timed
	 * out" messages to the tracerouting host for 15 seconds; as well,
	 * allow ICMP port unreachable from the tracrouted host for 15 secs
	 */
	if (ntohs(udp->dest) >= TRACEROUTE_MIN &&
	    ntohs(udp->dest) <= TRACEROUTE_MAX) {

		add_icmp_mask(0, 0, ip->ip_src, 0xFFFFFFFF, 11, 0, 15,
		    fw->interface);
		add_icmp(ip->ip_dst, ip->ip_src, 3, 3, 15, fw->interface);
		return(0);	/* don't need to allow the UDPs back in */
	}
	/*
	 * Allow return UDP packets
	 */
	
	debug(" %d %d", ntohs(udp->source), ntohs(udp->dest));
	/* install a reverse rule for 30 seconds */
	add_udp(ip->ip_dst, ip->ip_src, ntohs(udp->dest), ntohs(udp->source),
		30, fw->interface);

}

int
do_icmp(char *p) {

	struct ip *ip;
	struct icmphdr *icmp;
	struct fwhdr *fw=(struct fwhdr *)p;

	debug(" icmp");

	(char *)ip=p+20;
	(char *)icmp=(char *)ip+ip->ip_hl*4;
	/*
	 * Allow ICMP replies matching requests
	 */
	
	if (icmp->type==8 && icmp->code==0) {
		/* install a reverse rule for 30 seconds */

		debug(" echo request");
		add_icmp(ip->ip_dst, ip->ip_src, 0, 0, 30, fw->interface);
	} else
		debug(" (%d, %d)", icmp->type, icmp->code);

}

/* allow a rule for a packet from a src to a dest for a time limit */
add_rule(struct in_addr src, struct in_addr src_mask, struct in_addr dst,
	 struct in_addr dst_mask, int src_port_lo, int src_port_hi,
	 int dst_port_lo, int dst_port_hi, uint proto, time_t secs,
	 char *interface) {

	struct connection *conn;
	static struct ip_fwuser fw;
	int ret;

	/* update an existing entry if there is one */
	conn=NULL;
	if ((conn=find_connection(src, src_mask, dst, dst_mask, proto,
				  src_port_lo, src_port_hi, dst_port_lo,
				  dst_port_hi, interface))!=NULL) {
		conn->timeout=secs+time((time_t)NULL);
		return(0);
	}

	fw.ipfw.fw_src=src;
	fw.ipfw.fw_smsk=src_mask;
	fw.ipfw.fw_dst=dst;
	fw.ipfw.fw_dmsk=dst_mask;
	fw.ipfw.fw_proto=proto;
	fw.ipfw.fw_flg=0;
	fw.ipfw.fw_invflg=0;
	fw.ipfw.fw_spts[0]=src_port_lo;
	fw.ipfw.fw_spts[1]=src_port_hi;
	fw.ipfw.fw_dpts[0]=dst_port_lo;
	fw.ipfw.fw_dpts[1]=dst_port_hi;
	fw.ipfw.fw_redirpt=0;
	fw.ipfw.fw_outputsize=0;
	strcpy(fw.ipfw.fw_vianame, interface);
	strcpy(fw.label, "ACCEPT");
	nsaddrs=1;
	saddrs=&src;
	ndaddrs=1;
	daddrs=&dst;
	if ((ret=insert_entry("input", &fw, START_DYN_RULES, 0, 0))!=0) {
		fprintf(stderr, "error inserting a new input rule %s:%d\n",
			__FILE__, __LINE__);
	} else {
		record_connection(src, src_mask, dst, dst_mask, proto,
				  src_port_lo, src_port_hi, dst_port_lo,
				  dst_port_hi, secs, interface);	
	}

}

/* remove a rule */
delete_rule(struct connection *conn) {

	static struct ip_fwuser fw;
	int ret;

	memcpy(&fw.ipfw.fw_src, &conn->src_addr, sizeof(fw.ipfw.fw_src));
	memcpy(&fw.ipfw.fw_smsk, &conn->src_mask, sizeof(fw.ipfw.fw_smsk));
	memcpy(&fw.ipfw.fw_dst, &conn->dst_addr, sizeof(fw.ipfw.fw_dst));
	memcpy(&fw.ipfw.fw_dmsk, &conn->dst_mask, sizeof(fw.ipfw.fw_dmsk));
	fw.ipfw.fw_proto=conn->proto;
	fw.ipfw.fw_flg=0;
	fw.ipfw.fw_invflg=0;
	fw.ipfw.fw_spts[0]=conn->src_port_lo;
	fw.ipfw.fw_spts[1]=conn->src_port_hi;
	fw.ipfw.fw_dpts[0]=conn->dst_port_lo;
	fw.ipfw.fw_dpts[1]=conn->dst_port_hi;
	fw.ipfw.fw_redirpt=0;
	fw.ipfw.fw_outputsize=0;
	strcpy(fw.ipfw.fw_vianame, conn->interface);
	strcpy(fw.label, "ACCEPT");
	nsaddrs=1;
	saddrs=&conn->src_addr;
	ndaddrs=1;
	daddrs=&conn->dst_addr;
	if ((ret=delete_entry("input", &fw, 0, 0))!=0) {
		fprintf(stderr, "error deleteing an input rule %s:%d\n",
			__FILE__, __LINE__);
	} else {
		delete_connection(conn);	
	}

}

record_connection(struct in_addr src, struct in_addr src_mask,
		  struct in_addr dst, struct in_addr dst_mask,
		  uint proto, uint src_port_lo, uint src_port_hi,
		  uint dst_port_lo, uint dst_port_hi, time_t timeout,
		  char *interface) {

	struct connection *new_conn;

	if (debuglev>1) {

		char *ip_src, *ip_dst, *ip_src_mask, *ip_dst_mask;

		ip_src=strdup(inet_ntoa(src));
		ip_dst=strdup(inet_ntoa(dst));
		ip_src_mask=strdup(inet_ntoa(src_mask));
		ip_dst_mask=strdup(inet_ntoa(dst_mask));
		debug("recording connection %s:%s %s:%s %u->%u %u->%u %u %lu\n",
		      ip_src, ip_src_mask, ip_dst, ip_dst_mask,
		      proto, src_port_lo, src_port_hi, dst_port_lo,
		      dst_port_hi, timeout);
	}
	/* make timeout absolute */
	timeout+=time((time_t)NULL);

	if ((new_conn=(struct connection *)malloc(sizeof(struct connection)))==
	    NULL) {
		perror("Allocating new connection");
		exit(1);
	}
	if (!first_connection)
		first_connection=new_conn;

	memcpy(&new_conn->src_addr, &src, sizeof(src));
	memcpy(&new_conn->src_mask, &src_mask, sizeof(src_mask));
	memcpy(&new_conn->dst_addr, &dst, sizeof(dst));
	memcpy(&new_conn->dst_mask, &dst_mask, sizeof(dst_mask));
	new_conn->proto=proto;
	new_conn->src_port_lo=src_port_lo;
	new_conn->src_port_hi=src_port_hi;
	new_conn->dst_port_lo=dst_port_lo;
	new_conn->dst_port_hi=dst_port_hi;
	new_conn->timeout=timeout;
	new_conn->next=NULL;
	new_conn->prev=last_connection;
	strcpy(new_conn->interface, interface);
	if (last_connection)
		last_connection->next=new_conn;
	last_connection=new_conn;
	next_timeout=MIN(next_timeout, new_conn->timeout);

}

/* remove a connection from the table */
delete_connection(struct connection *conn){

	if (conn==first_connection)
		first_connection=conn->next;
	if (conn==last_connection)
		last_connection=conn->prev;

	if (conn->prev)
		conn->prev->next=conn->next;
	if (conn->next)
		conn->next->prev=conn->prev;

	free(conn);

}

display_connections() {

	struct connection *conn;
	char *ip_src, *ip_dst, *ip_src_mask, *ip_dst_mask;

	printf("connections table:\n");
	for (conn=first_connection; conn; conn=conn->next) {
		ip_src=strdup(inet_ntoa(conn->src_addr));
		ip_dst=strdup(inet_ntoa(conn->dst_addr));
		ip_src_mask=strdup(inet_ntoa(conn->src_mask));
		ip_dst_mask=strdup(inet_ntoa(conn->dst_mask));
		printf("%s:%s %s:%s %u->%u %u->%u %u %s %lu\n", ip_src,
		       ip_src_mask, ip_dst, ip_dst_mask, conn->proto,
		       conn->src_port_lo, conn->src_port_hi, conn->dst_port_lo,
		       conn->dst_port_hi, conn->interface, conn->timeout);

	}

}

expire_connections() {

	struct connection *conn, *next_conn;
	char *ip_src, *ip_dst, *ip_src_mask, *ip_dst_mask;
	time_t the_time;

	conn=first_connection;
	next_timeout=LONG_TIMEOUT;
	the_time=time((time_t)NULL);
	while(conn) {
		next_conn=conn->next;
		if (conn->timeout<=the_time) {
			if (debuglev>0) {
				ip_src=strdup(inet_ntoa(conn->src_addr));
				ip_dst=strdup(inet_ntoa(conn->dst_addr));
				ip_src_mask=strdup(inet_ntoa(conn->src_mask));
				ip_dst_mask=strdup(inet_ntoa(conn->dst_mask));
				debug("expired %s:%s %s:%s %u->%u %u->%u %u %lu\n",
				      ip_src, ip_src_mask, ip_dst, ip_dst_mask,
				      conn->proto, conn->src_port_lo,
				      conn->src_port_hi, conn->dst_port_lo,
				      conn->dst_port_hi, conn->timeout);
			}
			delete_rule(conn);
		} else {
			if (debuglev>0) {
				ip_src=strdup(inet_ntoa(conn->src_addr));
				ip_dst=strdup(inet_ntoa(conn->dst_addr));
				ip_src_mask=strdup(inet_ntoa(conn->src_mask));
				ip_dst_mask=strdup(inet_ntoa(conn->dst_mask));
				debug("not expired %s:%s %s:%s %u->%u %u->%u %u %lu\n",
				      ip_src, ip_src_mask, ip_dst, ip_dst_mask,
				      conn->proto, conn->src_port_lo,
				      conn->src_port_hi, conn->dst_port_lo,
				      conn->dst_port_hi, conn->timeout);
			}
			next_timeout=MIN(next_timeout, conn->timeout);
		}
		conn=next_conn;
	}
	if (next_timeout<=the_time)
		next_timeout=LONG_TIMEOUT;

}

struct connection *
find_connection(struct in_addr src, struct in_addr src_mask,
		  struct in_addr dst, struct in_addr dst_mask,
		  uint proto, uint src_port_lo, uint src_port_hi,
		  uint dst_port_lo, uint dst_port_hi, char *interface) {

	struct connection *conn, *next_conn;

	conn=first_connection;
	while(conn) {
		next_conn=conn->next;
		if (memcmp(&conn->src_addr, &src, sizeof(src))==0 &&
		    memcmp(&conn->src_mask, &src_mask, sizeof(src))==0 &&
		    memcmp(&conn->dst_addr, &dst, sizeof(dst))==0 &&
		    memcmp(&conn->dst_mask, &dst_mask, sizeof(dst))==0 &&
		    conn->proto==proto &&
		    conn->src_port_lo==src_port_lo &&
		    conn->src_port_hi==src_port_hi &&
		    conn->dst_port_lo==dst_port_lo &&
		    conn->dst_port_hi==dst_port_hi &&
		    strcmp(conn->interface, interface)==0) {
			return(conn);
		}
		conn=next_conn;
	}
	return((struct connection *)NULL);

}

install_spf_filters() {

	static struct ip_fwuser fw;
	int ret, c, line;
	char *v[100], buf[BUFSIZ];
	FILE *rules;

	/* flush input and output chains */
	if ((ret=flush_entries("input"))!=0) {
		fprintf(stderr, "error flushing input rules %s:%d\n",
			__FILE__, __LINE__);
	} 
	if ((ret=flush_entries("output"))!=0) {
		fprintf(stderr, "error flushing output rules %s:%d\n",
			__FILE__, __LINE__);
	} 

	/* install netlink filters on the output chain */
	fw.ipfw.fw_src.s_addr=0L;
	fw.ipfw.fw_smsk.s_addr=0L;
	fw.ipfw.fw_dst.s_addr=0L;
	fw.ipfw.fw_dmsk.s_addr=0L;
	fw.ipfw.fw_proto=IPPROTO_ICMP;
	fw.ipfw.fw_flg=IP_FW_F_NETLINK;
	fw.ipfw.fw_invflg=IP_FW_INV_VIA;
	fw.ipfw.fw_spts[0]=0;
	fw.ipfw.fw_spts[1]=0xFFFF;
	fw.ipfw.fw_dpts[0]=0;
	fw.ipfw.fw_dpts[1]=0xFFFF;
	fw.ipfw.fw_redirpt=0;
	fw.ipfw.fw_outputsize=0xFFFF;
	strcpy(fw.ipfw.fw_vianame, "lo");
	strcpy(fw.label, "ACCEPT");
	nsaddrs=1;
	saddrs=&fw.ipfw.fw_src;
	ndaddrs=1;
	daddrs=&fw.ipfw.fw_dst;
	if ((ret=insert_entry("output", &fw, 1, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 

	fw.ipfw.fw_proto=IPPROTO_UDP;
	if ((ret=insert_entry("output", &fw, 1, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 

	fw.ipfw.fw_proto=IPPROTO_TCP;
	if ((ret=insert_entry("output", &fw, 1, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 

	/* install default deny filter on the input chain */
	fw.ipfw.fw_src.s_addr=0L;
	fw.ipfw.fw_smsk.s_addr=0L;
	fw.ipfw.fw_dst.s_addr=0L;
	fw.ipfw.fw_dmsk.s_addr=0L;
	fw.ipfw.fw_proto=0;
	fw.ipfw.fw_flg=IP_FW_F_PRN;
	fw.ipfw.fw_invflg=0;
	fw.ipfw.fw_spts[0]=0;
	fw.ipfw.fw_spts[1]=0xFFFF;
	fw.ipfw.fw_dpts[0]=0;
	fw.ipfw.fw_dpts[1]=0xFFFF;
	fw.ipfw.fw_redirpt=0;
	fw.ipfw.fw_outputsize=0xFFFF;
	strcpy(fw.ipfw.fw_vianame, "");
	strcpy(fw.label, DEFAULT_ACTION);
	nsaddrs=1;
	saddrs=&fw.ipfw.fw_src;
	ndaddrs=1;
	daddrs=&fw.ipfw.fw_dst;
	if ((ret=insert_entry("input", &fw, 1, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 

#ifdef ALLOW_EST_TCP
	/* install a rule for established TCP on the input chain */
	fw.ipfw.fw_src.s_addr=0L;
	fw.ipfw.fw_smsk.s_addr=0L;
	fw.ipfw.fw_dst.s_addr=0L;
	fw.ipfw.fw_dmsk.s_addr=0L;
	fw.ipfw.fw_proto=IPPROTO_TCP;
	fw.ipfw.fw_flg=IP_FW_F_TCPSYN;
	fw.ipfw.fw_invflg=IP_FW_INV_SYN;
	fw.ipfw.fw_spts[0]=0;
	fw.ipfw.fw_spts[1]=0xFFFF;
	fw.ipfw.fw_dpts[0]=0;
	fw.ipfw.fw_dpts[1]=0xFFFF;
	fw.ipfw.fw_redirpt=0;
	fw.ipfw.fw_outputsize=0;
	strcpy(fw.ipfw.fw_vianame, "");
	strcpy(fw.label, "ACCEPT");
	nsaddrs=1;
	saddrs=&fw.ipfw.fw_src;
	ndaddrs=1;
	daddrs=&fw.ipfw.fw_dst;
	if ((ret=insert_entry("input", &fw, 1, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 
#endif

	/*
	 * install user defined rules in their own chain
	 */

	delete_chain("statinpt", 0);	/* try to remove the chain first */
	if ((ret=new_chain("statinpt"))==-1) {	/* create the new chain */
		perror("Creating statinpt chain");
		cleanup_exit(SIGTERM);
	}
	
	/* and link it as the last thing to do in the input chain */
	fw.ipfw.fw_src.s_addr=0L;
	fw.ipfw.fw_smsk.s_addr=0L;
	fw.ipfw.fw_dst.s_addr=0L;
	fw.ipfw.fw_dmsk.s_addr=0L;
	fw.ipfw.fw_proto=0;
	fw.ipfw.fw_flg=0;
	fw.ipfw.fw_invflg=0;
	fw.ipfw.fw_spts[0]=0;
	fw.ipfw.fw_spts[1]=0xFFFF;
	fw.ipfw.fw_dpts[0]=0;
	fw.ipfw.fw_dpts[1]=0xFFFF;
	fw.ipfw.fw_redirpt=0;
	fw.ipfw.fw_outputsize=0;
	strcpy(fw.ipfw.fw_vianame, "");
	strcpy(fw.label, "statinpt");
	nsaddrs=1;
	saddrs=&fw.ipfw.fw_src;
	ndaddrs=1;
	daddrs=&fw.ipfw.fw_dst;
	if ((ret=insert_entry("input", &fw, START_DYN_RULES, 0, 0))!=0) {
		fprintf(stderr, "error inserting netlink rule %s:%d\n",
			__FILE__, __LINE__);
	} 
	
	if ((rules=fopen("/etc/spf-rules", "r"))==NULL) {
		if (errno==ENOENT)
			return(0);
		else {
			perror("Opening rules file");
			cleanup_exit(SIGTERM);
		}
	}
	/* read a line at a time, vectorizing args into an array */
	c=0;
	v[c]="ipchains";
	line=0;
	while (!feof(rules)){
		
		char *s, *b;

		line++;
		if (fgets(buf, BUFSIZ, rules)==NULL)
			break;
		/* strip off comments */
		if (s=strchr(buf, '#'))
			*s='\0';
		if (s=strchr(buf, '\n'))
			*s='\0';	/* get rid of \n at the end */
		/* skip the line if it's blank */
		if (buf[0]=='\0')
			continue;
		b=buf;
		while (1) {
			v[++c]=b;
			if ((s=strchr(b, ' '))==NULL)
				break;
			while (*s==' ')
				*s++='\0';
			b=s;
			/* change any input chain specifiers to statinpt */
			if (strcmp(v[c], "input")==0)
				if (strcmp(v[c-1], "-A")==0 ||
				    strcmp(v[c-1], "-D")==0 ||
				    strcmp(v[c-1], "-R")==0 ||
				    strcmp(v[c-1], "-I")==0 ||
				    strcmp(v[c-1], "-P")==0)
					v[c]="statinpt";
		}
		/* check to see if we got anything at all */
		if (c==0)
			continue;	/* nope, next line please */
#if 0
		/* remove the \n from the last arg */
		v[c++][strlen(v[c])-1]='\0';
#else
		c++;
#endif
		/* run the installer */
		if (ipchains(c, v)!=0) {
			fprintf(stderr, "error installing rule on line %d\n",
				line);
			cleanup_exit(SIGTERM);
		}
		c=0;
	}
	fclose(rules);
	
}
